package dev.kske.chess.board;

import java.util.ArrayList;
import java.util.List;

import dev.kske.chess.board.Piece.Color;

/**
 * Project: <strong>Chess</strong><br>
 * File: <strong>Log.java</strong><br>
 * Created: <strong>09.07.2019</strong><br>
 * Author: <strong>Kai S. K. Engelbart</strong>
 */
public class Log implements Cloneable {

	private List<LoggedMove> moves;

	private Position	enPassant;
	private Color		activeColor;
	private int			fullmoveCounter, halfmoveClock;

	public Log() {
		moves = new ArrayList<>();
		reset();
	}

	@Override
	public Object clone() {
		Log log = null;
		try {
			log = (Log) super.clone();
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
		}
		log.moves = new ArrayList<>();
		log.moves.addAll(this.moves);
		return log;
	}

	/**
	 * Adds a move to the move history and adjusts the log to the new position.
	 * 
	 * @param move          The move to log
	 * @param capturedPiece The piece captured with the move
	 * @param pawnMove      {@code true} if the move was made by a pawn
	 */
	public void add(Move move, Piece capturedPiece, boolean pawnMove) {
		enPassant = pawnMove && move.yDist == 2 ? new Position(move.pos.x, move.pos.y + move.ySign) : null;
		if (activeColor == Color.BLACK) ++fullmoveCounter;
		if (pawnMove || capturedPiece != null) halfmoveClock = 0;
		else++halfmoveClock;
		activeColor = activeColor.opposite();
		moves.add(new LoggedMove(move, capturedPiece, enPassant, activeColor, fullmoveCounter, halfmoveClock));
	}

	/**
	 * @return the last logged move, or {@code null} if there is none
	 */
	public LoggedMove getLast() { return moves.isEmpty() ? null : moves.get(moves.size() - 1); }

	/**
	 * Removed the last move from the log and adjusts its state to the previous
	 * move.
	 */
	public void removeLast() {
		if (!isEmpty()) {
			moves.remove(moves.size() - 1);
			if (!isEmpty()) {
				LoggedMove last = moves.get(moves.size() - 1);
				activeColor		= last.activeColor;
				enPassant		= last.enPassant;
				fullmoveCounter	= last.fullmoveCounter;
				halfmoveClock	= last.halfmoveClock;
			} else reset();
		} else reset();
	}

	public boolean isEmpty() { return moves.isEmpty(); }

	/**
	 * Reverts the log to its initial state corresponding to the default board
	 * position.
	 */
	public void reset() {
		moves.clear();
		enPassant		= null;
		activeColor		= Color.WHITE;
		fullmoveCounter	= 1;
		halfmoveClock	= 0;
	}

	public List<LoggedMove> getLoggedMoves() { return moves; }

	public List<LoggedMove> getMoves() { return moves; }

	public void setMoves(List<LoggedMove> moves) { this.moves = moves; }

	public Position getEnPassant() { return enPassant; }

	public void setEnPassant(Position enPassant) { this.enPassant = enPassant; }

	public Color getActiveColor() { return activeColor; }

	public void setActiveColor(Color activeColor) { this.activeColor = activeColor; }

	public int getFullmoveCounter() { return fullmoveCounter; }

	public void setFullmoveCounter(int fullmoveCounter) { this.fullmoveCounter = fullmoveCounter; }

	public int getHalfmoveClock() { return halfmoveClock; }

	public void setHalfmoveClock(int halfmoveClock) { this.halfmoveClock = halfmoveClock; }

	public static class LoggedMove {

		public final Move		move;
		public final Piece		capturedPiece;
		public final Position	enPassant;
		public final Color		activeColor;
		public final int		fullmoveCounter, halfmoveClock;

		public LoggedMove(Move move, Piece capturedPiece, Position enPassant, Color activeColor, int fullmoveCounter,
				int halfmoveClock) {
			this.move				= move;
			this.capturedPiece		= capturedPiece;
			this.enPassant			= enPassant;
			this.activeColor		= activeColor;
			this.fullmoveCounter	= fullmoveCounter;
			this.halfmoveClock		= halfmoveClock;
		}
	}
}
